def person(name: str, age: int) -> None:
    """
    Enter person ID info.

    Parameters
    ----------
    name : str
        The person's first name.
    age : int
        The person's age.

    Returns
    -------
    None
        Nothing.

    Examples
    --------
    >>> from maite.testing.documentation.documentation_dependencies import person
    >>> person('Brad', 22)
    """

    return
