# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# ----
# Not exposing provider protocols publicly yet on the grounds that writing provider
# implementers will likely affect protocol design.
# ----
